/* ex1b.c <-- ex1b.rr, ex1b-2.rr 
   exp(xt-t^3) 
   $BF;>l$N#6>O$NNcBj(B. $BHyJ,J}Dx<0$O2<$N$h$&$KJQ$($F$^$9(B.
   gcc ex1b.c -lgsl -lblas -lm
*/
/*  [[p],[dt,q]] --> ( p - dt q ) f = 0.  
    (-3*dx^2+x) g - dt g = 0.
    (-3*dx^2+x) z - g(infty)+g(0)= 0.
    (-3*dx^2+x) z  = -1.  
    z0=z;
    z0' = z1,
    z1' = z0'' = (x/3)*z0 + (1/3)
    P=
    [[0,  1],
     [x/3,0]]
    exp(-x*T1+T3)*[z0,z1] $B$rHyJ,$9$k$H(B
     -T1*exp(-x*T1+T3)*[z0,z1]+exp(-x*T1+T3)*[z0,z1]'
   = -T1*exp(-x*T1+T3)*[z0,z1]+P*exp(-x*T1+T3)*[z0,z1]+exp(-x*T1+T3)*[0,1/3]
   = (-T*E + P)*exp(-x*T1+T3)*[z0,z1] + exp(-x*T1+T3)*[0,1/3]
  $B$^$H$a$F(B
   [y0,y1]' = (-T*E+P)*[y0,y1]+exp(-x*T1+T3)*[0,1/3]
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gsl/gsl_integration.h>
#define TMAX 100.0
double T1;
double T3;
     
double f (double t, void * params) {
  double f = exp(-t*t*t);
  return f;
}
double fd (double t, void * params) {
  double f = t*exp(-t*t*t);
  return f;
}
     
double
f0(void)
{
  gsl_integration_workspace * w 
        = gsl_integration_workspace_alloc (1000);
       
  double result, error;
     
  gsl_function F;
  F.function = &f;

     
  gsl_integration_qags (&F, 0, TMAX, 0, 1e-7, 1000,
                                                w, &result, &error); 
     
  printf ("result          = % .18f\n", result);
  printf ("estimated error = % .18f\n", error);
  printf ("intervals =  %d\n", w->size);
     
  return result;
}

double
f1(void)
{
  gsl_integration_workspace * w 
        = gsl_integration_workspace_alloc (1000);
       
  double result, error;
     
  gsl_function F;
  F.function = &fd;

     
  gsl_integration_qags (&F, 0, TMAX, 0, 1e-7, 1000,
                                                w, &result, &error); 
     
  printf ("result          = % .18f\n", result);
  printf ("estimated error = % .18f\n", error);
  printf ("intervals =  %d\n", w->size);
     
  return result;
}

#define RANK 2

double rkmin(double x0,double y0[],double xn,double minval[]);
main() {
  double y0[RANK];
  double minval[RANK];
  double x0,xn,minx;
  extern double T1,T3;
  T1 = 1.0; /* data average sum(t_i)/total */
  T3 = 1.0; /* data average sum(t_i^3)/total */
  x0 = 0.0; /* [x0,xn] is the search area */
  xn = 5.0;
  /* $B=i4|CM(B */
  y0[0] = f0()*exp(-0*T1+T3); 
  y0[1] = f1()*exp(-0*T1+T3);
  printf("Init y0=%lf\n", y0[0]);
  printf("Init y1=%lf\n", y0[1]);

  /* $B:G>.CM8!:w(B */
  minx=rkmin(x0,y0,xn,minval);
  printf("Opt x = %lf\n",minx);
}


/*
        Runge-Kutta
        y_{n+1} = y_n + 1/6 k_1 + 1/3 k_2 + 1/3 k_3 + 1/6 k_4
        k_1 = h f(x_n, y_n)
        k_2 = h f(x_n + 1/2 h, y_n + 1/2 k_1)
        k_3 = h f(x_n + 1/2 h, y_n + 1/2 k_2)
        k_4 = h f(x_n + h, y_n + k_3)
*/

/* $B%Y%/%H%kCM4X?t(B f(x, y) (x : scalar, y : vector) */ 
/* val : return value                              */
/*    P=
    [[0,  1],
     [x/3,0]]
    (-T1*E + P)*[Y0,Y1] + exp(-x*T1+T3)*[0,-1/3]
     [0,1/3] $B$X(B.
*/

void rf(double x, double *y, int m, double *val, int n)
{
  extern double T1;
  extern double T3;
  val[0] = -T1*y[1-1]+y[2-1];
  val[1] = (x/3)*y[1-1]-T1*y[2-1] + exp(-x*T1+T3)*(1.0/3.0);
}

void show_v(double *v, int n)
{
  int i;

  for (i = 0; i < n; i++)
    printf("%lf ", v[i]);
}

double rkmin(double x0,double y0[],double xn,double minval[])
{
  int i;
  double h = 0.01;
  double x; 
  double y[RANK];
  double k1[RANK], k2[RANK], k3[RANK], k4[RANK];
  double temp[RANK];
  double ty[RANK];
  double min;

  min = y0[0];
  for (i = 0; i < RANK; i++)
    y[i] = y0[i];
  for (x = x0; x < xn; x += h) {
    printf("%lf ", x);
    show_v(y, RANK); putchar('\n');

    rf(x, y, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k1[i] = h * temp[i];

    for (i = 0; i < RANK; i++)
      ty[i] = y[i] + 0.5 * k1[i];
    rf(x + 0.5 * h, ty, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k2[i] = h * temp[i];

    for (i = 0; i < RANK; i++)
      ty[i] = y[i] + 0.5 * k2[i];
    rf(x + 0.5 * h, ty, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k3[i] = h * temp[i];

    for (i = 0; i < RANK; i++)
      ty[i] = y[i] + k3[i];
    rf(x + h, ty, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k4[i] = h * temp[i];
                
    for (i = 0; i < RANK; i++)
      y[i] = y[i] + 1.0/6.0 * k1[i] + 1.0/3.0 * k2[i] + 1.0/3.0 * k3[i] + 1.0/6.0 * k4[i];
    if (y[0] < min) {
      min=y[0];
    }else{
      break;
    }
  }
  if (x < xn) {
    printf("Min at \n");
    printf("%lf ", x);
    show_v(y, RANK); putchar('\n');
  }else{
    printf("Min is not found in the domain [%lf,%lf]\n",x0,xn);
  }
  for (i=0; i<RANK; i++) minval[i]=y[i];
  return x;
}
