/* gcc gsl_matrix_test.c -lgsl -lblas on orange */
#include <stdio.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>

void set_lu(gsl_matrix *, gsl_matrix *, gsl_matrix *);
void test_set_lu();
void gsl_matrix_show(gsl_matrix *);
void test_gsl_show();

int main()
{
	double a_data[] = {1.0, 2.0, 3.0, 
	                   4.0, 5.0, 6.0,
			   7.0, 1.0, 9.0};
	double b_data[] = {1.0, 2.0, 5.0};

	gsl_matrix_view m = gsl_matrix_view_array(a_data, 3, 3);
	gsl_vector_view b = gsl_vector_view_array(b_data, 3);
	
	gsl_permutation *p = gsl_permutation_alloc(3);
	gsl_vector *x = gsl_vector_alloc(3);
	gsl_matrix *invm = gsl_matrix_alloc(3,3);
	gsl_matrix *c = gsl_matrix_alloc(3,3);
	gsl_matrix *mm = gsl_matrix_alloc(3,3);
	int s;

	gsl_matrix_show(&m.matrix);
	gsl_matrix_memcpy(mm, &m.matrix);
	gsl_linalg_LU_decomp(&m.matrix, p, &s);
	printf("LU decomp:\n");
	gsl_matrix_show(&m.matrix);
	gsl_linalg_LU_solve(&m.matrix, p, &b.vector, x);
	printf("x = \n");
	gsl_vector_fprintf(stdout, x, "%g");
	gsl_linalg_LU_invert(&m.matrix, p, invm);
	printf("inverse matrix:-----------\n");
	gsl_matrix_show(invm);
	printf("check inverse matrix:-----\n");
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, mm, invm, 0, c);
	gsl_matrix_show(c);

	printf("check LU decomp ----------\n");
	test_set_lu();

	gsl_permutation_free(p);
	gsl_vector_free(x);
	gsl_matrix_free(invm);
	gsl_matrix_free(c);
	gsl_matrix_free(mm);
	return 0;
}


/* m is the output of gsl_linalg_LU_decomp */
void set_lu(gsl_matrix *m, gsl_matrix *l, gsl_matrix *u)
{
	int i, j;

	for (i = 0; i < m->size1; i++)
		gsl_matrix_set(l, i, i, 1.0);
	for (i = 0; i < m->size1; i++) {
		for (j = 0; j < m->size2; j++) {
			if (i <= j)
				gsl_matrix_set(u, i, j, gsl_matrix_get(m, i, j));
			else 
				gsl_matrix_set(l, i, j, gsl_matrix_get(m, i, j));
		}
	}
}

void test_set_lu()
{
	double a_data[] = {1.0, 2.0, 3.0, 
	                   4.0, 5.0, 6.0,
			   7.0, 1.0, 9.0};
	double b_data[] = {1.0, 2.0, 5.0};

	gsl_matrix_view m = gsl_matrix_view_array(a_data, 3, 3);
	gsl_vector_view b = gsl_vector_view_array(b_data, 3);
	gsl_permutation *p = gsl_permutation_alloc(3);
	int s;
	gsl_matrix *l = gsl_matrix_alloc(3,3);
	gsl_matrix *u = gsl_matrix_alloc(3,3);
	gsl_matrix *c = gsl_matrix_alloc(3,3);


	gsl_matrix_show(&m.matrix);
	gsl_linalg_LU_decomp(&m.matrix, p, &s);
	printf("LU decomp:\n");
	gsl_matrix_show(&m.matrix);
	set_lu(&m.matrix, l, u);
	printf("L:\n");
	gsl_matrix_show(l);
	printf("U:\n");
	gsl_matrix_show(u);

	/* d : double, ge : general, mm : matrix mul */
	/* c <- 1.0 * l * u + 0.0 * c                */
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, l, u, 0.0, c);
	printf("P * L * U =\n");
	gsl_matrix_show(c);

	gsl_permutation_free(p);
	gsl_matrix_free(l);
	gsl_matrix_free(u);
	gsl_matrix_free(c);
}

void gsl_matrix_show(gsl_matrix *mat)
{
	int i, j, m, n;

	m = mat->size1;
	n = mat->size2;
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) 
			printf("%8f ", gsl_matrix_get(mat, i, j));
		printf("\n");
	}
}

void test_gsl_matrix_show()
{
	gsl_matrix *m;

	m = gsl_matrix_alloc(3,3);
	gsl_matrix_set_identity(m);
	gsl_matrix_show(m);
}
