/* From GSL Manual */
/* gcc gsl_multimin_test.c -lgsl -lblas */
#include <stdio.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_multimin.h>


/* ¿δؿ f(x,y) = c*(x-a)^2 + d*(y-b)^2 + e  (a,b,c,d,e : params) */
double my_f(const gsl_vector *v, void *params)
{
	double x, y;
	double *p = (double *)params;

	x = gsl_vector_get(v, 0);
	y = gsl_vector_get(v, 1);
printf("call my_f(%lf, %lf)\n", x, y);
	return p[2] * (x - p[0]) * (x - p[0]) + p[3] * (y - p[1]) * (y - p[1]) + p[4];
}

/* f θ df = (df/dx, df/dy) */
void my_df(const gsl_vector *v, void *params, gsl_vector *df)
{
	double x, y;
	double *p = (double *)params;

	x = gsl_vector_get(v, 0);
	y = gsl_vector_get(v, 1);

printf("call my_df(%lf, %lf)\n", x, y);
	gsl_vector_set(df, 0, 2.0 * p[2] * (x - p[0]));
	gsl_vector_set(df, 1, 2.0 * p[3] * (y - p[1]));
}

/* f  df ξ֤ؿ */
void my_fdf(const gsl_vector *x, void *params, double *f, gsl_vector *df)
{
	*f = my_f(x, params);
	my_df(x, params, df);
}

int main (void)
{
	size_t iter = 0;
	int status;
	const gsl_multimin_fdfminimizer_type *T;
	gsl_multimin_fdfminimizer *s;

	/* 濴(1,2)(10,20)Ǿͤ30 */
	double par[5] = { 1.0, 2.0, 10.0, 20.0, 30.0 };

	gsl_vector *x;
	/* ŪؿΥǡ */
	gsl_multimin_function_fdf my_func;
	my_func.n = 2;          /* õ֤μ */
	my_func.f = &my_f;      /* Ūؿ       */
	my_func.df = &my_df;    /* ۥ٥ȥ   */
	my_func.fdf = &my_fdf;  /* Ūؿȸۥ٥ȥξ׻ؿ */
	my_func.params = &par;  /* ѥ᡼     */

	/* õx = (5, 7) */
	x = gsl_vector_alloc(2);
	gsl_vector_set(x, 0, 5.0);
	gsl_vector_set(x, 1, 7.0);

#ifdef FR 
	/* Fletcher-Reeves ˤ붦ˡ (FR ˡ) */
	T = gsl_multimin_fdfminimizer_conjugate_fr;
#else
	/* ǵ޹߲ˡ ޤǽɤʤǥ */
	T = gsl_multimin_fdfminimizer_steepest_descent;
#endif
	s = gsl_multimin_fdfminimizer_alloc(T, 2);
	/* 
	 * ؿ my_func õ x õ
	 * 0.01  step size
	 * 1e-4 ľõ
	 */
/*	gsl_multimin_fdfminimizer_set (s, &my_func, x, 0.01, 1e-4); */
	gsl_multimin_fdfminimizer_set (s, &my_func, x, 0.01, 0.1);
	do {
		iter++;
		status = gsl_multimin_fdfminimizer_iterate(s);
		if (status) break;
		/*  s->gradient ΥΥȽꡣ1e-3 ̤ˤʤ GSL_SUCCESS ֤ */
		status = gsl_multimin_test_gradient(s->gradient, 1e-3);
		if (status == GSL_SUCCESS) printf("Minimum found at:\n");
		printf("%5d %.5f %.5f %10.5f\n", iter, gsl_vector_get (s->x, 0), gsl_vector_get (s->x, 1), s->f);
	} while (status == GSL_CONTINUE && iter < 100);
	gsl_multimin_fdfminimizer_free(s);
	gsl_vector_free(x);
	return 0;
}
