with(LinearAlgebra):
with(ListTools):

subsrule:=proc(L,V)
  local n,i,R;
  n:=nops(L);
  R:=[];
  for i from 1 to n do
    R := [op(R),L[i]=V[i]]:
  od:
end proc:

mysubs:=proc(E,L,V)
  local Rule;
  Rule:=subsrule(L,V);
  evalf(subs(Rule,E));
end proc:

runge_kutta_replace_linear:=proc(V,F,Y,N,X,IND,Xk,Rule_vector,Sign)
  local V1,i, Pf, GS;
  Pf := pfaffian(evalf(subs(X=Xk,F)),X,Xk,IND,F[4],Sign);
#  V1:=evalf(subs(X=Xk,F));
  V1:=Pf.Rule_vector;
  for i from 1 to N do
    V[i] := V1[i];
  od:
end proc:

runge_kutta_4_linear:=proc(F,X,IND,Y,X0,Y0,H,CX1,OneStep,Sign)
  local K1,K2,K3,K4;
  local Ans,N,Yk,Yk1,Xk,X1;
  X1:=CX1;

  #OneStep = getopt(onestep);
  # if (type(OneStep) <= 0) OneStep = 0; else OneStep = 1;
  if (OneStep > 0) then X1:=X0+2*H end if;
  if ( (H<0) and (X1-X0)>0 ) then
    error("taka_runge_kutta_4_linear, X1-X0 should be <0")
  end if:
  if ((H>0) and (X1-X0)<0) then
    error("taka_runge_kutta_4_linear, X1-X0 should be >0")
  end if;
  Ans := [];
  if (X0 = X1) then return([[X0,op(Y0)]]) end if;

  N := RowDimension(F[2][1][1]);

  Yk := Y0;
  K1 := Vector(N);
  K2 := Vector(N);
  K3 := Vector(N);
  K4 := Vector(N);
  Yk1 := Vector(N);
  Xk := X0;

  while (condop(H < 0, Xk > X1, Xk < X1)) do
    runge_kutta_replace_linear(K1,F,Y,N,X,IND,Xk,Yk,Sign);
    runge_kutta_replace_linear(K2,F,Y,N,X,IND,Xk+H*(1/2),Yk+K1*(1/2)*H,Sign);
    runge_kutta_replace_linear(K3,F,Y,N,X,IND,Xk+H*(1/2),Yk+K2*(1/2)*H,Sign);
    runge_kutta_replace_linear(K4,F,Y,N,X,IND,Xk+H,Yk+K3*H,Sign);
    Yk1 := Yk+H*(K1/6+K2/3+K3/3+K4/6);
#print(K1,K2,K3,K4,Yk1);
 #   if (Taka_Runge_kutta_debug) print([Xk,Yk[1]]);
 #   if (Taka_Runge_kutta_save_data) {
      Ans := [[Xk,op(convert(Yk,`list`))],op(Ans)];
 #   }
    if (OneStep > 0) then
       return([[Xk+H,op(convert(Yk1,'list'))], [Xk,op(convert(Yk,'list'))]]);
    end if;
 #   if (Taka_Runge_kutta_graphic0) glib_line(Xk,Yk[1],Xk+H,Yk1[1]);
 #   if (Taka_Runge_kutta_adapted0 && 
 #       (taka_runge_kutta_abs(Yk1-Yk) > Taka_Runge_kutta_epsilon)) {
 #     H = H*(1/2);
 #   }else{
 #     if (Taka_Runge_kutta_adapted0) H = H*2;
      Xk := Xk+H;
      Yk := Yk1;
 #   }
  od:
  return(Ans);
end proc:

condop:=proc(Con,A,B)
  if Con then A else B end if;
end proc:

runge_kutta_4a_linear:=proc(FF,X0,IND,Y,S0,CYs,T0,H,OneStep,Sign)
  local A,B,T,Ys;
  
  Ys := CYs;
  if ( type(Ys,list) ) then Ys := Vector(Ys); end if;
  
  if (T0 < S0) then
 #   /* opposite direction */
    return runge_kutta_reverse(runge_kutta_4a_linear(subs(X0=-X0,FF),X0,IND,Y,-S0,Ys,-T0,H,0,-1));
  end if;
  if (H >= T0-S0) then
    A:=runge_kutta_4_linear(FF,X0,IND,Y,S0,Ys,T0-S0,0,1,Sign); # | onestep=1);
  else
    A:=runge_kutta_4_linear(FF,X0,IND,Y,S0,Ys,H,T0,0,Sign);
    T:=A[1];
    if (T0-T[1] > 0) then
      B:=runge_kutta_4_linear(FF,X0,IND,Y,T[1],Vector(subsop(1=NULL,T)),T0-T[1],0,1,Sign);# | onestep=1);
      T:=B[1];
      A:=[T,op(A)];
    end if;
  end if;
  return(A);
end proc:

runge_kutta_reverse:=proc(CA)
  local A, B, T;
  A := CA;
  B := [];
  while (not nops(A) = 0) do
    T := A[1];
    B := [op(B),[-T[1],op(subsop(1=NULL,T))]];
    A := subsop(1=NULL,A);
  od:
  return(B);
end proc:



## for gradient method

runge_kutta_4a_linear_grad:=proc(FF,X0,IND,Y,S0,CYs,T0,H,OneStep,Sign,GRAD)
  local A,B,T,Ys;
  
  Ys := CYs;
  if ( type(Ys,list) ) then Ys := Vector(Ys); end if;
  
  if (T0 < S0) then
 #   /* opposite direction */
    return runge_kutta_reverse(runge_kutta_4a_linear_grad(subs(X0=-X0,FF),X0,IND,Y,-S0,Ys,-T0,H,0,-1,GRAD));
  end if;
  if (H >= T0-S0) then
    A:=runge_kutta_4_linear_grad(FF,X0,IND,Y,S0,Ys,T0-S0,0,1,Sign,GRAD); # | onestep=1);
  else
    A:=runge_kutta_4_linear_grad(FF,X0,IND,Y,S0,Ys,H,T0,0,Sign,GRAD);
    T:=A[1];
    if (T0-T[1] > 0) then
      B:=runge_kutta_4_linear_grad(FF,X0,IND,Y,T[1],Vector(subsop(1=NULL,T)),T0-T[1],0,1,Sign,GRAD);# | onestep=1);
      T:=B[1];
      A:=[T,op(A)];
    end if;
  end if;
  return(A);
end proc:

runge_kutta_replace_linear_grad:=proc(V,F,Y,N,X,IND,Xk,Rule_vector,Sign,GRAD)
  local V1,i, Pf, GS, TF, GPF;

  GPF := Matrix(op(1,Y),op(1,Y));
  TF := evalf(subs(X=Xk,F));
  for i from 1 to nops(IND[1]) do
    Pf := pfaffian(TF,IND[1][i],Xk,IND[2][i],F[4],Sign);
#print("Pf",IND[1][i],Pf);
    GPF := GPF + GRAD[i]*Pf;
  od;
#  V1:=evalf(subs(X=Xk,F));
  V1:=GPF.Rule_vector;
  for i from 1 to N do
    V[i] := V1[i];
  od:
end proc:

runge_kutta_4_linear_grad:=proc(F,X,IND,Y,X0,Y0,H,CX1,OneStep,Sign,GRAD)
  local K1,K2,K3,K4;
  local Ans,N,Yk,Yk1,Xk,X1;
  X1:=CX1;

  #OneStep = getopt(onestep);
  # if (type(OneStep) <= 0) OneStep = 0; else OneStep = 1;
  if (OneStep > 0) then X1:=X0+2*H end if;
  if ( (H<0) and (X1-X0)>0 ) then
    error("taka_runge_kutta_4_linear, X1-X0 should be <0")
  end if:
  if ((H>0) and (X1-X0)<0) then
    error("taka_runge_kutta_4_linear, X1-X0 should be >0")
  end if;
  Ans := [];
  if (X0 = X1) then return([[X0,op(Y0)]]) end if;

  N := RowDimension(F[2][1][1]);

  Yk := Y0;
  K1 := Vector(N);
  K2 := Vector(N);
  K3 := Vector(N);
  K4 := Vector(N);
  Yk1 := Vector(N);
  Xk := X0;

  while (condop(H < 0, Xk > X1, Xk < X1)) do
    runge_kutta_replace_linear_grad(K1,F,Y,N,X,IND,Xk,Yk,Sign,GRAD);
    runge_kutta_replace_linear_grad(K2,F,Y,N,X,IND,Xk+H*(1/2),Yk+K1*(1/2)*H,Sign,GRAD);
    runge_kutta_replace_linear_grad(K3,F,Y,N,X,IND,Xk+H*(1/2),Yk+K2*(1/2)*H,Sign,GRAD);
    runge_kutta_replace_linear_grad(K4,F,Y,N,X,IND,Xk+H,Yk+K3*H,Sign,GRAD);
    Yk1 := Yk+H*(K1/6+K2/3+K3/3+K4/6);
#print(K1,K2,K3,K4,Yk1);
 #   if (Taka_Runge_kutta_debug) print([Xk,Yk[1]]);
 #   if (Taka_Runge_kutta_save_data) {
      Ans := [[Xk,op(convert(Yk,`list`))],op(Ans)];
 #   }
    if (OneStep > 0) then
       return([[Xk+H,op(convert(Yk1,'list'))], [Xk,op(convert(Yk,'list'))]]);
    end if;
 #   if (Taka_Runge_kutta_graphic0) glib_line(Xk,Yk[1],Xk+H,Yk1[1]);
 #   if (Taka_Runge_kutta_adapted0 && 
 #       (taka_runge_kutta_abs(Yk1-Yk) > Taka_Runge_kutta_epsilon)) {
 #     H = H*(1/2);
 #   }else{
 #     if (Taka_Runge_kutta_adapted0) H = H*2;
      Xk := Xk+H;
      Yk := Yk1;
 #   }
  od:
  return(Ans);
end proc:

