#include "test_h.h"

double g_x11, g_x12, g_x22, g_y1, g_y2, g_r;

void test_rk(); 
int rf(double y1, const double *y, double *val, void *params);
void invmat(gsl_matrix *m, gsl_matrix *invm);
void gsl_matrix_show(gsl_matrix *mat);

void test_rk()
{
	/*            x11 , x12, x22 , y1 , y2  , r  */ 
	double p[] = {-3.0, 0.5, 0.15, 1.0, 0.15, 1.0};
	double x11,x12,x22,y1,y2,r;
	x11 = -3.0; x12 = 0.5; x22 = 0.15; 
	y1 = 1.0; y2 = 0.15; r = 1.0;

	/* initial value */
	double val[] = 
{2.948670364423784504,0.5817985920578351965,0.4754127542095007750,0.6566284657983176130};


extern gsl_matrix *a1, *inv_a1, *b1, *c1, *e1; 
extern gsl_matrix *a2, *inv_a2, *b2, *c2, *e2; 
extern gsl_matrix *p2, *inv_p2, *q2; 
extern gsl_matrix *p3, *inv_p3, *q3, *r3; 
extern gsl_matrix *pf1_m;

extern  double g_x11, g_x12, g_x22, g_y1, g_y2, g_r;
	g_x11 = -3.0; g_x12 = 0.5; g_x22 = 0.15; 
	g_y2 = 0.15; g_r = 1.0;
	y1 = 1.0;

	/* Runge-Kutta-Fehlberg(4,5) */
	const gsl_odeiv_step_type *T = gsl_odeiv_step_rkf45;
	gsl_odeiv_step *s = gsl_odeiv_step_alloc(T, 4);
	gsl_odeiv_control *c = gsl_odeiv_control_y_new(1e-6, 0.0);
	gsl_odeiv_evolve *e = gsl_odeiv_evolve_alloc(4);
	gsl_odeiv_system sys = {rf, NULL, 4, NULL};

	double h = 1e-6; /* ƥåν */
	double goal = 1.5;

	while (y1 < goal) {
		int status = gsl_odeiv_evolve_apply(e, c, s, &sys, &y1, goal, &h, val);
		if (status != GSL_SUCCESS) break;
		printf("%f %f %f %f %f\n", y1, val[0], val[1], val[2], val[3]);
	}

	gsl_odeiv_evolve_free(e);
	gsl_odeiv_control_free(c);
	gsl_odeiv_step_free(s);

}

int rf(double y1, const double *y, double *val, void *params)
{
extern gsl_matrix *pf1_m;
extern double g_x11, g_x12, g_x22, g_y2, g_r;
	int i, j;

	pf1(g_x11, g_x12, g_x22, y1, g_y2, g_r);

	for (i = 0; i < 4; i++) {
		val[i] = 0.0;
		for (j = 0; j < 4; j++)
			val[i] += gsl_matrix_get(pf1_m, i, j) * y[j];
	}

	return GSL_SUCCESS;
}

void invmat(gsl_matrix *m, gsl_matrix *invm)
{
	int n = m->size1;
	int s;
	gsl_permutation *p = gsl_permutation_alloc(n);

	gsl_linalg_LU_decomp(m, p, &s);
	gsl_linalg_LU_invert(m, p, invm);

	gsl_permutation_free(p);
}

void test_invmat()
{
	gsl_matrix *invm = gsl_matrix_alloc(3,3);	
	gsl_matrix *c = gsl_matrix_alloc(3,3);	
	gsl_matrix *mm = gsl_matrix_alloc(3,3);	
	gsl_matrix *mmm = gsl_matrix_alloc(3,3);	

	gsl_matrix_set(mm, 0, 0, 1.0);
	gsl_matrix_set(mm, 0, 1, 2.0);
	gsl_matrix_set(mm, 0, 2, 3.0);
	gsl_matrix_set(mm, 1, 0, 4.0);
	gsl_matrix_set(mm, 1, 2, 5.0);
	gsl_matrix_set(mm, 1, 2, 6.0);
	gsl_matrix_set(mm, 2, 0, 7.0);
	gsl_matrix_set(mm, 2, 1, 1.0);
	gsl_matrix_set(mm, 2, 2, 9.0);
	gsl_matrix_memcpy(mmm, mm);
	invmat(mm, invm);
	gsl_matrix_show(invm);
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, mmm, invm, 0.0, c);
	gsl_matrix_show(c);

	gsl_matrix_free(invm);
	gsl_matrix_free(c);
	gsl_matrix_free(mm);
	gsl_matrix_free(mmm);
}

void gsl_matrix_show(gsl_matrix *mat)
{
	int i, j, m, n;

	m = mat->size1;
	n = mat->size2;
	for (i = 0; i < m; i++) {
		for (j = 0; j < n; j++) 
			printf("%8f ", gsl_matrix_get(mat, i, j));
		printf("\n");
	}
}

